try:
    import bsread
except:
    bsread = None

from datahub import *
from datahub.utils.checker import check_msg
import collections
import threading

_logger = logging.getLogger(__name__)

class Bsread(Source):
    """
    Retrieves data from a Bsread source.
    """

    DEFAULT_URL = os.environ.get("BSREAD_DEFAULT_URL", None if (bsread is None) else bsread.DEFAULT_DISPATCHER_URL)

    def __init__(self, url=DEFAULT_URL, mode="SUB", path=None, **kwargs):
        """
        url (str, optional): Stream URL. Default value can be set by the env var BSREAD_DEFAULT_URL.
        mode (str, optional): "SUB" or "PULL"
        path (str, optional): hint for the source location in storage or displaying.
        """
        Source.__init__(self, url=url, path=path, **kwargs)
        if bsread is None:
            raise Exception("BSREAD library not available")
        self.mode = mode
        self.context = 0

    def run(self, query):
        mode = bsread.PULL if self.mode == "PULL" else bsread.SUB
        receive_timeout = query.get("receive_timeout", 3000)
        channels = query.get("channels", None)
        filter = query.get("filter", None)
        if not self.url or (self.url == bsread.DEFAULT_DISPATCHER_URL):
            host, port = None, 9999
            stream_channels = channels
        else:
            host, port = get_host_port_from_stream_address(self.url)
            stream_channels = None

        self.context = None

        with bsread.source(host=host, port=port, mode=mode, receive_timeout=receive_timeout, channels=stream_channels) as stream:
            self.context = stream.stream.context
            while not self.range.has_ended() and not self.aborted:
                data = stream.receive()
                if not data:
                    raise Exception("Received None message.")
                if self.range.has_started():
                    timestamp = create_timestamp(data.data.global_timestamp, data.data.global_timestamp_offset)
                    pulse_id = data.data.pulse_id
                    format_changed = data.data.format_changed
                    data=data.data.data

                    keys = channels if (channels and (len(channels)>0)) else data.keys()
                    msg = {channel: data[channel].value for channel in keys}

                    try:
                        if not filter or self.is_valid(filter, id, timestamp, msg):
                            self.on_msg(pulse_id, timestamp, msg, format_changed)
                    except Exception as e:
                        _logger.exception("Error receiving data: %s " % str(e))

            self.close_channels()
        if self.context:
            self.context.destroy()
            self.context = None

    def is_valid(self, filter, id, timestamp, msg):
        try:
            return check_msg(msg, filter)
        except Exception as e:
            _logger.warning("Error processing filter: %s " % str(e))
            return False

    def on_msg(self, id, timestamp, msg, format_changed):
        for channel in msg.keys():
            try:
                self.receive_channel(channel, msg[channel], timestamp, id, check_changes=format_changed)
            except Exception as ex:
                _logger.exception("Error receiving data: %s " % str(ex))


    def close(self):
        if self.context:
            self.context.destroy()
            self.context = None
        Source.close(self)


class BsreadStream(Bsread):

    def __init__(self, channels, filter=None, queue_size=100,  **kwargs):
        Bsread.__init__(self, **kwargs)
        self.message_buffer = collections.deque(maxlen=queue_size)
        self.condition = threading.Condition()
        self.req(channels, 0.0, 365 * 24 * 60 * 60, filter=filter, background=True)

    def close(self):
        Bsread.close(self)

    def on_msg(self, id, timestamp, msg, format_changed):
        with self.condition:
            self.message_buffer.append((id, timestamp, msg))
            self.condition.notify()

    def drain(self):
        with self.condition:
            self.message_buffer.clear()

    def receive(self, timeout=None):
        with self.condition:
            if not self.message_buffer:
                self.condition.wait(timeout)
            if self.message_buffer:
                return self.message_buffer.popleft()




import unittest
from datahub import *

url = "tcp://localhost:5554"
filename = "/Users/gobbo_a/tst.h5"
channels = ["intensity", "height"]
start = None
end = 1.0
query = {
    "channels": channels,
    "start": start,
    "end": end,
    "mode": "SUB"
}


url_pipeline_server = "http://localhost:8889"
url_camera_server = "http://localhost:8888"
pipeline = "simulation_sp"
camera = "simulation"

class CamserverTest(unittest.TestCase):
    def setUp(self):
        self.source = Bsread(url=url)
        self.pipeline = Pipeline( url=url_pipeline_server, name=pipeline)
        self.camera = Camera(url=url_camera_server, name=camera)

    def tearDown(self):
        cleanup()

    def test_bsread(self):
        hdf5 = HDF5Writer(filename)
        #stdout = Stdout()
        table = Table()
        self.source.add_listener(hdf5)
        #self.source.add_listener(stdout)
        self.source.add_listener(table)
        self.source.request(query)
        dataframe = table.as_dataframe()
        self.assertEqual(list(dataframe.keys()), channels)

    def test_pipeline(self):
        #hdf5 = HDF5Writer(filename)
        stdout = Stdout()
        table = Table()
        #self.pipeline.add_listener(hdf5)
        self.pipeline.add_listener(table)
        self.pipeline.add_listener(stdout)
        self.pipeline.request(query)
        dataframe = table.as_dataframe()
        self.assertEqual(list(dataframe.keys()), channels)

    def test_camera(self):
        query["channels"] = None
        hdf5 = HDF5Writer(filename)
        table = Table()
        self.camera.add_listener(hdf5)
        self.camera.add_listener(table)
        self.camera.request(query)
        dataframe = table.as_dataframe()
        self.assertEqual(list(dataframe.keys()), ['width', 'height', 'image', 'x_axis', 'y_axis'])
        self.camera.close_listeners()

    def test_config(self):
        with Pipeline(url=url_pipeline_server, name="[simulation3_sp]", config = {"binning_x":2,"binning_y":2}) as source:
            stdout = Stdout()
            source.add_listener(stdout)
            source.req(start = 0.0, end=2.0, channels = ["width", "height"])

        config = {
            "pipeline_type": "processing",
            "camera_name": "simulation"
        }
        with Pipeline(url=url_pipeline_server, config=config) as source:
            stdout = Stdout()
            source.add_listener(stdout)
            source.request(query)

        with Pipeline(url=url_pipeline_server, name="tst", config=config) as source:
            stdout = Stdout()
            source.add_listener(stdout)
            source.request(query)

        with Pipeline(url=url_pipeline_server, name=pipeline) as source:
            stdout = Stdout()
            source.add_listener(stdout)
            source.request(query)

        with Pipeline(url=url_pipeline_server, name="test[simulation3_sp]", config = {"binning_x":2, "binning_y":2}) as source:
            stdout = Stdout()
            source.add_listener(stdout)
            source.req(start = 0.0, end=2.0, channels = ["width", "height"])


if __name__ == '__main__':
    unittest.main()



from datahub.sources.bsread import Bsread

class Dispatcher(Bsread):
    """
    Retrieves data from the DataBuffer dispatcher.
    """

    def __init__(self, path=None, **kwargs):
        """
        path (str, optional): hint for the source location in storage or displaying.
        """
        Bsread.__init__(self, url=None, mode="SUB", path=path, **kwargs)



import unittest
from datahub.main import run_json


json_str = '{' \
           '"file": "/Users/gobbo_a/dev/back/json.h5", ' \
           '"print": true, ' \
           '"epics": {"url": null, "query":{"start":null, "end":3.0, "channels": ["TESTIOC:TESTCALCOUT:Input", "TESTIOC:TESTSINUS:SinCalc", "TESTIOC:TESTWF2:MyWF"]}},' \
           '"bsread": {"url": "tcp://localhost:9999", "mode":"PULL", "query":{"start":null, "end":3.0, "channels":  ["UInt8Scalar", "Float32Scalar"]}}' \
           '}'

json_str = '{' \
           '"file": "/Users/gobbo_a/dev/back/json.h5", ' \
           '"print": true, ' \
           '"start": null, ' \
           '"end": 3.0, ' \
           '"epics": {"channels": ["TESTIOC:TESTCALCOUT:Input", "TESTIOC:TESTSINUS:SinCalc", "TESTIOC:TESTWF2:MyWF"]},' \
           '"bsread": {"url": "tcp://localhost:9999", "mode":"PULL", "channels":  ["UInt8Scalar", "Float32Scalar"]}' \
           '}'

class EpicsTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_json(self):
        run_json(json_str)


if __name__ == '__main__':
    unittest.main()



import os
import pathlib
from setuptools import setup, find_packages

PACKAGE_PREFIX = "psi-"
PACKAGE_NAME = "datahub"

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def version():
    p = pathlib.Path(__file__).parent.joinpath(PACKAGE_NAME).joinpath("package_version.txt")
    with open(p, "r") as f1:
        return f1.read()[:-1]


setup(
    name=PACKAGE_PREFIX + PACKAGE_NAME,
    version=version(),
    packages=find_packages(),
    author="Paul Scherrer Institute",
    author_email="daq@psi.ch",
    description="Utilities to retrieve data from PSI sources.",
    license="GPLv3",
    keywords="",
    url="https://github.com/paulscherrerinstitute/" + PACKAGE_NAME,
    long_description=read('Readme.md'),
    long_description_content_type="text/markdown",
    install_requires=[
    ],
    entry_points={
        'console_scripts': [
            f'{PACKAGE_NAME} = {PACKAGE_NAME}.main:main',
        ],
    },
    data_files=[
        (f"{PACKAGE_NAME}", [f"{PACKAGE_NAME}/package_version.txt"])
    ]
)


import dash
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output
import plotly.graph_objects as go
import numpy as np

# Generate example data
x = np.linspace(0, 10, 500)
y = np.sin(x)
y_upper_error = 0.1 + 0.2 * np.sqrt(x)
y_lower_error = 0.1 + 0.1 * np.sqrt(x)

# Create the Dash app
app = dash.Dash(__name__)
c = (0 ,0, 255)
# Define the layout of the app
app.layout = html.Div(children=[
    html.H1(children='Line Plot with Error Bars'),

    dcc.Graph(
        id='line-error-plot',
        figure={
            'data': [
                go.Scatter(
                    x=x,
                    y=y,
                    mode='lines+markers',
                    name='Data',
                    line=dict(color=f'rgba({c[0]}, {c[1]}, {c[2]}, 1.0)'),
                    marker=dict(
                        color=f'rgba({c[0]}, {c[1]}, {c[2]}, 1.0)',
                        size=5
                    ),
                    error_y=dict(
                        type='data',
                        symmetric=False,
                        array=y_upper_error,
                        arrayminus=y_lower_error,
                        visible=True,
                        color=f'rgba({c[0]}, {c[1]}, {c[2]}, 0.1)'  # Set the color with transparency
                    )
                )
            ],
            'layout': go.Layout(
                title='Line Plot with Asymmetric Error Bars and Transparency',
                xaxis={'title': 'X Axis'},
                yaxis={'title': 'Y Axis'}
            )
        }
    )
])

# Run the app
if __name__ == '__main__':
    app.run_server(debug=True)



import unittest
from datahub import *

class RetrivalTest(unittest.TestCase):
    def setUp(self):
        self.retrieval = Retrieval()
        #self.databuffer = DataBuffer(url="https://data-api.psi.ch/sf")
        self.databuffer = DataBuffer()

    def tearDown(self):
        self.retrieval.close()
        self.databuffer.close()


    def test_search_retrieval(self):
        self.retrieval.search("SARFE10-PSSS059")


    def test_search_databuffer(self):
        self.databuffer.print_search("SARFE10-PSSS059")

    def test_search_daqbuf(self):
        with Daqbuf(backend=None) as source:
            source.print_search("FPICTURE")

if __name__ == '__main__':
    unittest.main()




from datahub import *


with Daqbuf(backend="sf-databuffer", cbor=True, parallel=True) as source:
    table = Table()
    source.add_listener(table)

    source.req("SAT-CVME-TIFALL5:EvtSet", 21928758437, 21928758467)
    print(len(table.data["SAT-CVME-TIFALL5:EvtSet"]))
    dataframe = table.as_dataframe(Table.PULSE_ID)
    print(dataframe)

    table.close()
    source.req("SAT-CVME-TIFALL5:EvtSet", -1.0, 0.0)
    print(len(table.data["SAT-CVME-TIFALL5:EvtSet"]))
    dataframe = table.as_dataframe(Table.TIMESTAMP)
    print(dataframe)


    source.req("SAT-CVME-TIFALL5:EvtSet", -11.0, -9.0)
    print(len(table.data["SAT-CVME-TIFALL5:EvtSet"]))
    dataframe = table.as_dataframe(Table.TIMESTAMP)
    print(dataframe)
    table.clear()
    source.req("SAT-CVME-TIFALL5:EvtSet", "2024-06-26 16:00:00",
               "2024-06-26 16:19:00")
    print(len(table.data["SAT-CVME-TIFALL5:EvtSet"]))
    dataframe = table.as_dataframe(Table.TIMESTAMP)
    print(dataframe)


import time
import unittest
from datahub import *


class PlotTest(unittest.TestCase):

    def test_epics(self):
        with Plot() as plot:
            with Epics(url="localhost:54321", time_type="str") as source:
                source.add_listener(plot)
                source.req(["TESTIOC:TESTSINUS:SinCalc", "TESTIOC:TESTSINUS2:SinCalc"], 0.0, 4.0)

    def test_camera(self):
        with Plot(max_rate=2.0,) as plot:
             with Camera(url="http://localhost:8888", name="simulation") as source:
                source.add_listener(plot)
                source.req(None, 0.0, 2.0)


    def test_data_buffer(self):
        with Plot() as plot:
             with DataBuffer(time_type="str", ) as source:
                source.add_listener(plot)
                source.req( ["SARFE10-PSSS059:SPECTRUM_Y", "SARFE10-PSSS059:SPECTRUM_X"], "2024-02-14 08:50:00.000", "2024-02-14 08:50:05.000")


    def test_bsread(self):
        with Plot() as plot:
             with Bsread(url="tcp://localhost:9999", mode="PULL", time_type="str", ) as source:
                source.add_listener(plot)
                source.req( ["UInt8Scalar", "Float32Scalar"], 0.0, 2.0, modulo=50)

if __name__ == '__main__':
    unittest.main()



from datahub import *

_logger = logging.getLogger(__name__)

KNOWN_BACKENDS = ["sf-databuffer", "sf-archiverappliance"]

class DataBuffer(Source):
    """
    Retrieves data from the DataBuffer.
    """

    DEFAULT_URL = os.environ.get("DATA_BUFFER_DEFAULT_URL", "https://data-api.psi.ch/sf-databuffer")
    DEFAULT_BACKEND = os.environ.get("DATA_BUFFER_DEFAULT_BACKEND", "sf-databuffer")

    def __init__(self, url=DEFAULT_URL, backend=DEFAULT_BACKEND, path=None, delay=1.0, **kwargs):
        """
        url (str, optional): DataBuffer URL. Default value can be set by the env var DATA_BUFFER_DEFAULT_URL.
        backend (str, optional): DataBuffer backend. Default value can be set by the env var DATA_BUFFER_DEFAULT_BACKEND.
        path (str, optional): hint for the source location in storage or displaying.
        delay (float, optional): Wait time for channels to be uploaded to storage before retrieval.
        """
        if url is None:
            raise RuntimeError("Invalid URL")
        Source.__init__(self, url=url, backend=backend, query_path="/query",  search_path="/channels", path=path,
                        known_backends=KNOWN_BACKENDS, **kwargs)
        self.delay = delay

    def _get_range(self, start_expansion=False, end_expansion=False):
        type = self.range.end_type
        if type == "id":
            start = self.range.get_start_id()
            end = self.range.get_end_id()
            return {"endPulseId": str(end), "startPulseId": str(start),
                    "startExpansion": start_expansion, "endExpansion": end_expansion}
        elif type == "date":
            start = self.range.get_start_str_iso()
            end = self.range.get_end_str_iso()
            #return {"startDate": datetime.datetime.isoformat(start), "endDate": datetime.datetime.isoformat(end),
            #        "startExpansion": start_expansion, "endExpansion": end_expansion}
            return {"startDate": start, "endDate": end,
                    "startExpansion": start_expansion, "endExpansion": end_expansion}
        elif type == "time":
            start = self.range.get_start_sec()
            end = self.range.get_end_sec()
            return {"startSeconds": "%.9f" % start, "endSeconds": "%.9f" % end,
                    "startExpansion": start_expansion, "endExpansion": end_expansion}
        else:
            raise Exception("Undefined query type")

    def run(self, query):
        import requests
        channels = query.get("channels",[])
        start_expansion = query.get("start_expansion", False)   #Expand query range on start until next datapoint (can be a very expensive operation depending on the backend)
        end_expansion = query.get("end_expansion", False)       # Expand query range on end until next datapoint (can be a very expensive operation depending on the backend
        aggregation = query.get("aggregation", None)
        server_side_mapping = query.get("server_side_mapping", None)
        server_side_mapping_strategy = query.get("server_side_mapping_strategy", "provide-as-is")

        self.range.wait_end(delay=self.delay)

        if isinstance(channels, str):
            channels = [channels, ]
        # Build up channel list for the query
        channel_list = []
        for channel in channels:
            channel_name = channel.split("/")

            if len(channel_name) > 2:
                raise RuntimeError("%s is not a valid channel specification" % channel)
            elif len(channel_name) == 1:
                channel_list.append({"name": channel_name[0]})
            else:
                channel_list.append({"name": channel_name[1], "backend": channel_name[0]})


        query = dict()
        query["channels"] = channel_list
        #query["fields"] = ["pulseId", "globalSeconds", "globalDate", "value", "eventCount"]
        query["fields"] = ["pulseId", "globalSeconds", "globalDate", "value"]

        # Set query ranges
        query["range"] = self._get_range( start_expansion=start_expansion, end_expansion=end_expansion)
        # Set aggregation
        if aggregation is not None:
            query["aggregation"] = aggregation.get_json()

        if server_side_mapping:
            query["mapping"] = {"incomplete": server_side_mapping_strategy}

        response = requests.post(self.url, json=query)

        # Check for successful return of data
        if response.status_code != 200:
            raise RuntimeError("Unable to retrieve data from server: ", response)

        data = response.json()

        for ch in data:
            check_changes=True
            name = ch["channel"]["name"]
            for rec in ch["data"]:
                timestamp = create_timestamp(float(rec["globalSeconds"]))   #rec['eventCount']
                pulse_id = rec["pulseId"]
                value = rec["value"]
                self.receive_channel(name, value, timestamp,pulse_id, check_changes=check_changes, check_types=True)
                check_changes = False
        self.close_channels()

    def search(self, regex):
        import requests
        cfg = {
            "regex": regex,
            "ordering": "asc",
            "reload": "true"
        }
        if self.backend is not None:
            cfg["backends"] = [self.backend]
        response = requests.post(self.search_url, json=cfg)
        ret = response.json()
        if not self.verbose:
            pd = self._get_pandas()
            if pd is None:
                if self.backend and ret[0]['backend'] == self.backend:
                    ret = ret[0]['channels']
            else:
                if len(ret) == 0:
                    return None
                data = []
                for src in ret:
                    for channel in src["channels"]:
                        data.append([src["backend"], channel])
                df = pd.DataFrame(data, columns=["backend", "name"])
                df = df.sort_values(by=["backend", "name"])
                columns_to_display = ["backend", "name"]
                ret = df[columns_to_display].to_string(index=False)
        return ret




import time
import datetime
import logging
from datetime import datetime, timezone

_logger = logging.getLogger(__name__)
try:
    from dateutil import parser as dateutil_parser
except:
    _logger.error("dateutil not installed: fewer data formats are supported")
    dateutil_parser=None

PULSE_ID_START_TIME = 1504951960.114 # This  is fluctuating and changed after Jul/24 from 1504531686.91
PULSE_ID_INTERVAL = 0.01
PULSE_ID_INTERVAL_DEC = len(str(PULSE_ID_INTERVAL).split('.')[1]) if '.' in str(PULSE_ID_INTERVAL) else 0

def create_timestamp(sec, nano=0):
    #Doing in 2 steps, because if multiply by 10e9 that there will rounding error that will change the final timestamp
    micros = int(sec * 1000000)
    return (micros * 1000) + nano

def convert_timestamp(timestamp, type="nano"):
    if type == "str":
        secs = float(timestamp) / 1000000000.0
        return timestamp_to_string(secs, False)[:-3]
    elif type =="sec":
        return float(timestamp) / 1000000000.0
    elif type =="milli":
        return int(timestamp/1000000)
    return timestamp

def time_to_pulse_id(tm=time.time()):
    tm = tm - get_utc_offset()
    offset = tm - PULSE_ID_START_TIME
    id = int(offset / PULSE_ID_INTERVAL)
    return id

def pulse_id_to_time(id):
    offset = float(id) * PULSE_ID_INTERVAL
    ret = PULSE_ID_START_TIME + offset
    ret = ret + get_utc_offset()
    return round(ret, PULSE_ID_INTERVAL_DEC)


def timestamp_to_string(seconds=time.time(), utc=True):
    if utc:
        dt = datetime.utcfromtimestamp(seconds)
        return dt.strftime('%Y-%m-%d %H:%M:%S.%f') + 'Z'
    else:
        dt = datetime.fromtimestamp(seconds)
        return dt.strftime('%Y-%m-%d %H:%M:%S.%f')

def string_to_timestamp(date_string):
    dt = string_to_datetime(date_string)
    return dt.timestamp()

def string_to_datetime(date_string):
    if dateutil_parser:
        return dateutil_parser.parse(date_string)
    else:
        return datetime.fromisoformat(date_string.rstrip('Z')).replace(tzinfo=timezone.utc)

def get_utc_offset():
    now = time.time()
    dt_utc = datetime.utcfromtimestamp(now)
    dt_local = datetime.fromtimestamp(now)
    difference = dt_local-dt_utc
    return difference.total_seconds()


import re

CHANNEL_NAME_ALLOWED_SYMBOLS =  ":-"
CHANNEL_NAME_PATTERN = "[\w" + CHANNEL_NAME_ALLOWED_SYMBOLS + "]"

def check_msg_single_statement(msg, statement):
    pattern = r"^(?P<name>" + CHANNEL_NAME_PATTERN + "+)(?P<op>==|!=|<=|>=|>|<)(?P<value>.+)$"
    match = re.match(pattern, statement)
    if not match:
        raise ValueError(f"Invalid statement: {statement}")
    name = match.group("name")
    op = match.group("op")
    value_str = match.group("value")

    if name not in msg:
        return False
    actual_value = msg[name]

    #Convert the value to the appropriate type (int, float, bool, or str)
    try:
        if value_str.lower() == 'true':
            value = True
        elif value_str.lower() == 'false':
            value = False
        elif value_str.isdigit():
            value = int(value_str)
        else:
            try:
                value = float(value_str)
            except ValueError:
                value = value_str.strip('"\'')  # Remove surrounding quotes for string
    except ValueError:
        raise ValueError(f"Invalid value: {value}")

    # Try using eval to perform the comparison
    #try:
    #    return eval(f"{repr(actual_value)} {op} {repr(value)}")
    #except Exception:
    #    return False
    # Perform the comparison with no eval to improve performance
    if op == "==":
        return actual_value == value
    elif op == "!=":
        return actual_value != value
    elif op == "<=":
        return actual_value <= value
    elif op == ">=":
        return actual_value >= value
    elif op == "<":
        return actual_value < value
    elif op == ">":
        return actual_value > value
    else:
        raise ValueError(f"Invalid operator: {op}")


def check_msg(msg, statement):
    def tokenize(expression):
        #token_pattern = re.compile(r'\s*(AND|OR|\(|\)|True|False|\w+(<|>|[!=<>]=)[^\s()]+)\s*')
        token_pattern = re.compile(r'\s*(AND|OR|\(|\)|True|False|' + CHANNEL_NAME_PATTERN + '+(<|>|[!=<>]=)[^\s()]+)\s*')
        tokens = token_pattern.findall(expression)
        # Extract the first element from each tuple in tokens, which contains the actual token.
        tokens = [token[0] for token in tokens]
        return tokens

    def eval_expr(entry, tokens):
        def parse_value():
            token = tokens.pop(0)
            if token == '(':
                val = eval_expr(entry, tokens)
                if tokens and tokens[0] == ')':
                    tokens.pop(0)  # Remove ')'
                return val
            elif token == 'True':
                return True
            elif token == 'False':
                return False
            elif re.match(r"^" + CHANNEL_NAME_PATTERN + "+([!=<>]=|>|<).+$", token):
                return check_msg_single_statement(entry, token)
            else:
                raise ValueError(f"Unexpected token: {token}")

        values = [parse_value()]

        while tokens:
            if not tokens:
                break
            op = tokens.pop(0)
            if op == ')':
                break
            next_value = parse_value()
            if op == 'AND':
                values[-1] = values[-1] and next_value
            elif op == 'OR':
                values[-1] = values[-1] or next_value
            else:
                raise ValueError(f"Unexpected operator: {op}")

        return values[0]

    tokens = tokenize(statement)

    # Add a check for balanced parentheses
    if tokens.count('(') != tokens.count(')'):
        raise ValueError("Unbalanced parentheses in the statement")

    if len(tokens) == 1: #Speed up normal case (1 simple statement)
        return check_msg_single_statement(msg, tokens[0])
    return eval_expr(msg, tokens)


import logging
import sys
import traceback

import numpy as np
from datahub import Consumer
from datahub.utils.timing import string_to_timestamp, convert_timestamp
import multiprocessing
import time

_logger = logging.getLogger(__name__)


plt = None
def import_plplot():
    global plt
    try:
        import matplotlib.pyplot
        plt =  matplotlib.pyplot
    except:
        raise Exception("Cannot import matplotlib")

plots = {}


def is_notebook():
    return plt.get_backend() in ["nbAgg", "backend_interagg", "module://backend_interagg"]

def create_plot(name, shape, typ, xdata, start, colormap, color, marker_size, line_width, max_count):
    try:
        global plots
        if len(shape) == 0:
            fig, ax = plt.subplots(num=name)
            ax.set_title(name)
            plot, = ax.plot([], [], marker='o', color=color, markersize=marker_size, linewidth=line_width)
            # ax.set_xlim(0, 10)
            # ax.set_ylim(0, 100)
        elif len(shape) == 1:
            fig, ax = plt.subplots(num=name)
            ax.set_title(name)
            plot, = ax.plot([], [], marker='.', color=color)
            plot.set_xdata(xdata)
        elif len(shape) == 2:
            image_data = np.zeros(shape, typ)
            fig, ax = plt.subplots(num=name)
            plot = ax.imshow(image_data, cmap=colormap, origin='lower')
        else:
            return
        if not is_notebook():
            plt.ion()  # Turn on interactive mode
            plt.show()  # Display the current subplot


        plots[name] = (plot, shape, fig, ax, max_count)
    except Exception as ex:
        print(f"Exception creating {name}: {str(ex)}")
        traceback.print_exc()


def update_plot(name, timestamp, value):
    try:
        if name in plots:
            (plot, shape, fig, ax, max_count) = plots[name]
            if plot is not None:
                if len(shape) == 0:
                    x, y = np.append(plot.get_xdata(), timestamp), np.append(plot.get_ydata(), value)
                    if max_count:
                        if len(x) > max_count:
                            x, y = x[-max_count:], y[-max_count:]

                    plot.set_xdata(x)
                    plot.set_ydata(y)
                    ax.relim()  # Recalculate the data limits
                    ax.autoscale_view()  # Autoscale the view based on the data limits
                elif len(shape) == 1:
                    plot.set_ydata(value)
                    ax.relim()  # Recalculate the data limits
                    ax.autoscale_view()  # Autoscale the view based on the data limits
                else:
                    plot.set_array(value)
                    plot.norm = plt.Normalize(value.min(), value.max())
                    plt.draw()
            if not is_notebook():
                repaint(0.01)

    except Exception as ex:
        print(f"Exception adding to {name}: {str(ex)}")
        traceback.print_exc()

def show_plot(name):
    if name in plots:
        del plots[name]
        if len(plots) == 0:
            try:
                #plt.ioff()  # Turn off interactive mode
                #plt.show(block=False)
                pass
            except:
                pass

def get_open_figures():
    open_figures = [fig for fig in plt._pylab_helpers.Gcf.get_all_fig_managers()]
    return len(open_figures)


def process_plotting(tx_queue,  stop_event):
    _logger.info("Start plotting process")
    stop_event.clear()
    import_plplot()
    try:
        while not stop_event.is_set():
            try:
                tx= tx_queue.get(False)
            except:
                repaint(0.01)
                continue
            if tx is not None:
                if tx[0] == "START":
                    create_plot(*tx[1:])
                elif tx[0] == "REC":
                    update_plot(*tx[1:])
                elif tx[0] == "END":
                    show_plot(*tx[1:])
    except Exception as e:
        _logger.exception(e)
    finally:
        stop_event.set()
        if is_notebook():
            plt.show()
        else:
            while get_open_figures() > 0:
                repaint()

        _logger.info("Exit plotting process")
        sys.exit(0)


def repaint(interval=0.1, show_plot=False):
    if get_open_figures() > 0:
        #plt.pause(interval)

        #Doing this instead of calling pause in order multiple windows not to fliker
        manager = plt.get_current_fig_manager()
        if manager is not None:
            canvas = manager.canvas
            if canvas.figure.stale:
                canvas.draw_idle()
            if show_plot:
               plt.show(block=False)
            canvas.start_event_loop(interval)
            return
    time.sleep(interval)

class Plot(Consumer):
    def __init__(self,  channels=None, colormap="viridis", color=None,  marker_size=None, line_width=None, max_count=None, max_rate=None, **kwargs):
        Consumer.__init__(self, **kwargs)
        self.plots = {}
        self.channels = channels
        self.min_interval = (1.0/max_rate) if max_rate else None
        self.last_plotted_record={}
        self.colormap = colormap.lower() if colormap else None
        self.color = color
        self.marker_size= marker_size
        self.line_width = line_width
        self.max_count = int(max_count) if max_count else None

        import_plplot()

        self.tx_queue = multiprocessing.Queue()
        self.stop_event = multiprocessing.Event()
        self.stop_event.set()
        self.plotting_process = multiprocessing.Process(target=process_plotting, args=(self.tx_queue, self.stop_event))
        self.plotting_process.start()
        #Wait process to start
        start = time.time()
        while self.stop_event.is_set():
            if time.time() - start > 5.0:
                raise Exception("Cannot start plotting process")
            time.sleep(0.01)


    def on_close(self):
        while not self.tx_queue.empty():
            time.sleep(0.1)
        self.stop_event.set()
        self.plots = {}
        self.tx_queue.close()


    def on_start(self, source):
        pass

    def on_stop(self, source, exception):
        pass

    def on_channel_header(self, source, name, typ, byteOrder, shape, channel_compression, metadata):
        if self.channels:
            if not name in self.channels:
                return
        if len(shape)==0:
            xdata = None
        elif len(shape) == 1:
            xdata = list(range(shape[0]))
        elif len(shape) == 2:
            xdata = None
        else:
            _logger.warning("Unsupported shape for channel: " + name)
            return
        self.tx_queue.put(["START", name, shape, typ, xdata, time.time(), self.colormap, self.color, self.marker_size, self.line_width, self.max_count])
        self.plots[name] = [shape, xdata, time.time()]
        time.sleep(0.1)

    def on_channel_record(self, source, name, timestamp, pulse_id, value, **kwargs):
        if name in self.plots:
            shape, xdata, start = self.plots[name]
            try:
                if self.min_interval:
                    if len(shape) > 0: #Only downsample arrays
                        timespan = time.time() - self.last_plotted_record.get(name, 0.0)
                        if timespan < self.min_interval:
                            return
                if type(timestamp) == str:
                    timestamp = string_to_timestamp(timestamp)
                if type(timestamp) == int:
                    timestamp = float(timestamp)/10e9
                if isinstance(value, np.floating):  # Different scalar float types don't change header
                    value = float(value)
                if isinstance(value, np.integer):  # Different scalar int types don't change header
                    value = int(value)
                timestamp = timestamp - start
                self.tx_queue.put(["REC", name, timestamp, value])
                if self.min_interval:
                    self.last_plotted_record[name] = time.time()
                #time.sleep(0.1)
            except Exception as e:
                print("Error in plotting: " + str(e))



    def on_channel_completed(self, source, name):
        self.tx_queue.put(["END", name])







import time
import unittest
from datahub import *

filename = "/Users/gobbo_a/dev/back/mult.h5"
channels_epics = ["TESTIOC:TESTCALCOUT:Input", "TESTIOC:TESTSINUS:SinCalc", "TESTIOC:TESTWF2:MyWF"]
url_bsread = "tcp://localhost:9999"
mode_bsread= "PULL"
channels_bsread = ["UInt8Scalar", "Float32Scalar"]
start = None
end = 3.0


class EpicsTest(unittest.TestCase):
    def setUp(self):
        self.epics = Epics()
        self.bsread = Bsread(url=url_bsread, mode=mode_bsread)


    def tearDown(self):
        cleanup()


    def test_listeners(self):
        hdf5 = HDF5Writer(filename)
        stdout = Stdout()
        table = Table()
        self.epics.add_listener(hdf5)
        self.epics.add_listener(stdout)
        self.epics.add_listener(table)
        self.bsread.add_listener(hdf5)
        self.bsread.add_listener(stdout)
        self.bsread.add_listener(table)

        self.epics.req(channels_epics, start, end, background=True)
        self.bsread.req(channels_bsread, start, end, background=True)
        #time.sleep(1.0)
        #self.bsread.abort()
        self.epics.join()
        self.bsread.join()
        dataframe = table.as_dataframe()
        print (dataframe.columns)


if __name__ == '__main__':
    unittest.main()



import unittest
from datahub import *


class RangeTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_range(self):
        query = {}
        query["start"] = "2024-01-29 11:48:00Z"
        query["end"] = "2024-01-29 12:48:00.050Z"
        now = time.time()

        # print (now)

        tr = QueryRange(query)
        t = time.time()
        print(t)

        id = tr.time_to_id(t)
        print(id)

        tm = tr.id_to_time(id)
        print(tm)

        self.assertLess(abs(tm - t), PULSE_ID_INTERVAL)

        print(tr.seconds_to_string(tm, False))

        print(tr.seconds_to_string(now, False))

        """
        def time_to_id(self, time=datetime.now()):
            start_pulse_id = datetime(2017, 9, 4, 11, 11, 18)
            millis = int((now - start_pulse_id).total_seconds() * 1000)
            pid = millis // 10
            return pid
        """
        """
        #print(tr.time_to_id())
        s1=tr.seconds_to_string(now, True)
        s2=tr.seconds_to_string(now, False)
        print (s1)
        print (s2)

        s1 = tr.string_to_seconds(s1)
        s2 = tr.string_to_seconds(s2)
        print (s1)
        print (s2)

        print (tr.get_start())
        print (tr.get_end())
        print (tr.get_start_str())
        print (tr.get_end_str())

        """



if __name__ == '__main__':
    unittest.main()




import dash
from dash import dcc, html, Output, Input

app = dash.Dash(__name__)

app.layout = html.Div([
    dcc.Store(id='page-load-event'),
    html.Div(id='output')
])

# Client-side callback to trigger on page load
app.clientside_callback(
    """
    function(n_intervals) {
        return n_intervals;
    }
    """,
    Output('page-load-event', 'data'),
    Input('interval', 'n_intervals'),
    prevent_initial_call=False
)

# Interval component to trigger the client-side callback
app.layout.children.append(dcc.Interval(id='interval', interval=1, n_intervals=0))

if __name__ == '__main__':
    app.run_server(debug=True)



import logging
import numpy as np
from datahub import Consumer
from datahub.utils.timing import timestamp_to_string

_logger = logging.getLogger(__name__)

class TextColors:
    RESET = "\033[0m"
    RED = "\033[91m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    BLUE = "\033[94m"
    MAGENTA = "\033[95m"
    CYAN = "\033[96m"

class Stdout(Consumer):
    def __init__(self,  **kwargs):
        Consumer.__init__(self, **kwargs)
        self.align = "{:<32} {:<24} {:<12} {:<40}"
        self.columns = ("Name", "Timestamp", "ID", "Value")
        self.color_header = TextColors.CYAN
        self.color_sources = TextColors.BLUE
        self.color_names = TextColors.RED
        self.width = 120
        self.first_record = True

    def center(self, message):
        message_width = len(message)
        if (message_width + 2)< self.width:
            padding = (self.width - message_width) // 2 - 1
            message = '-' * padding + " " + message + " " + '-' * (padding)
        return message

    def print_header(self):
        print(f"{self.color_header}" + self.align.format(*self.columns) + f"{TextColors.RESET}")

    def on_close(self):
        pass

    def on_start(self, source):
        now = timestamp_to_string(utc=False)
        msg = self.center(f"Starting {source.get_desc()} at {now}")
        print(f"{self.color_sources}{msg}{TextColors.RESET}")
        msg = self.center(f"{source.query}")
        print(f"{self.color_sources}{msg}{TextColors.RESET}")

    def on_stop(self, source, exception):
        now = timestamp_to_string(utc=False)
        msg = self.center(f"Finished {source.get_desc()} at {now}")
        print(f"{self.color_sources}{msg}{TextColors.RESET}")

    def on_channel_header(self, source, name, typ, byteOrder, shape, channel_compression, metadata):
        #print (f"{self.color_names}\t+{name} - {typ} {byteOrder} {shape} {channel_compression}{TextColors.RESET}")
        pass

    def on_channel_record(self, source, name, timestamp, pulse_id, value, **kwargs):
        if self.first_record:
            self.print_header()
            self.first_record = False
        if type(value) == bytes:
            value  = f"bytes({len(value)})"
        elif isinstance(value, np.ndarray) and (len(value.shape) >= 1):
            tokens = str(value).split('\n')
            val_str = tokens[0]
            if len(tokens) > 1:
                val_str = val_str + " ..."
            value = f"{str(value.dtype)}{str(value.shape)} {val_str}"
        else:
            value = str(value)

        if kwargs.get("bins", None):
            value = f"{value}  \t [{kwargs['min']} - {kwargs['max']}]"
        print(self.align.format(
            str(name),
            str(timestamp),
            str(pulse_id),
            value
        ))

    def on_channel_completed(self, source, name):
        #print (f"{self.color_names}\t-{name}{TextColors.RESET}")
        pass




import sys
import zmq
import json
import base64
import threading


class PlotClient:
    """
    A Python client to PShell plot server
    """

    def __init__(self, address="localhost", port=7777, context=None, timeout=None):
        self.set_context(context)
        self.context = zmq.Context()
        self.url = "tcp://" + address + (":%s" % port)
        self.debug = False
        self.timeout = None if (timeout is None) else int(timeout * 1000)
        self._create_socket()
        self.lock = threading.Lock()

    def _create_socket(self):
        if self.debug: print("Connecting to server: " + self.url)
        self.socket = self.context.socket(zmq.REQ)
        if self.timeout is not None:
            self.socket.setsockopt(zmq.LINGER, 0)
            self.poller = zmq.Poller()
            self.poller.register(self.socket, zmq.POLLIN)
        self.socket.connect(self.url)

    def execute(self, cmd_name, cmd_data):
        with self.lock:
            if self.debug: print("Sending " + cmd_name + "...")
            json_str = json.dumps(cmd_data)
            self.socket.send_string(cmd_name, flags=zmq.SNDMORE)
            self.socket.send_string(json_str)
            if self.debug: print("Receiving...")

            if not self.timeout is None:
                if not self.poller.poll(self.timeout):
                    self.socket.close()
                    self._create_socket()
                    raise TimeoutError("Communication timeout")
            rx = self.socket.recv_string()
            if self.debug: print("Done")
            response = json.loads(rx)
            error = response["error"]
            if (error is not None) and (len(error) > 0):
                raise Exception(error)
            return response["ret"]

    def close(self):
        if self.debug: print("Closing...")
        self.socket.close()
        self.context.term()

    def set_context(self, context):
        self.ctx = context if context else "0"

    def get_contexts(self):
        ret = self.execute("GetContexts", {})
        ret = json.loads(ret)
        return ret

    def set_context_attrs(self, quality=None, layout=None):
        """
            quality (str): Low, Medium, High or Maximum
            layout (str): Vertical, Horizontal, Grid
        """
        self.execute("SetContextAttrs", {"context": self.ctx, "quality": quality, "plotLayout": layout})

    def clear_contexts(self):
        self.execute("ClearContexts", {})

    def clear_plots(self):
        self.execute("ClearPlots", {"context": self.ctx})

    def add_line_plot(self, title=None, style=None):
        """
            style = Normal (default), Step, Spline, ErrorX, ErrorY or ErrorXY
        """
        return self.execute("AddLinePlot", {"context": self.ctx, "title": title, "style": style})

    def add_matrix_plot(self, title=None, style=None, colormap=None):
        """
            colormap = Grayscale, Red, Green, Blue, Inverted, Temperature (Default), Rainbow, Flame
            style = Normal (Default), Mesh, Image.
        """
        return self.execute("AddMatrixPlot",
                            {"context": self.ctx, "title": title, "style": style, "colormap": colormap})

    def add_3d_plot(self, title=None, colormap=None):
        """
            colormap = Grayscale, Red, Green, Blue, Inverted, Temperature (Default), Rainbow, Flame
            style = Normal (Default), Mesh, Image.
        """
        return self.execute("Add3dPlot", {"context": self.ctx, "title": title, "colormap": colormap})

    def add_time_plot(self, title=None, started=None, duration=None, markers=None):
        return self.execute("AddTimePlot",
                            {"context": self.ctx, "title": title, "started": started, "duration": duration,
                             "markers": markers})

    def add_table(self, title=None):
        return self.execute("AddTable", {"context": self.ctx, "title": title})

    def set_progress(self, progress):
        """
        0.0 < progress < 1.0 , or None for hiding it
        """
        self.execute("SetProgress", {"context": self.ctx, "progress": progress})

    def set_status(self, status):
        self.execute("SetStatus", {"context": self.ctx, "status": status})

    def set_table_data(self, table, header, data):
        """
        header: list with the column titles
        data: list of lists holding the table contents
        """
        self.execute("SetTableData", {"table": table, "header": header, "data": data})

    def clear_plot(self, plot):
        self.execute("ClearPlot", {"plot": plot})

    def clear_series(self, series):
        self.execute("ClearSeries", {"series": series})

    def set_line_plot_attrs(self, plot, style):
        """
            style: see add_line_plot
        """
        self.execute("SetLinePlotAttrs", {"plot": plot, "style": style})

    def set_matrix_plot_attrs(self, plot, colormap):
        """
            colormap : see add_matrix_plot
        """
        self.execute("SetMatrixPlotAttrs", {"plot": plot, "colormap": colormap})

    def set_time_plot_attrs(self, plot, started=None, duration=None, markers=None):
        """
            colormap : see add_matrix_plot
        """
        self.execute("SetTimePlotAttrs", {"plot": plot, "started": started, "duration": duration, "markers": markers})

    def set_plot_axis_attrs(self, axis, label=None, auto_range=None, min=None, max=None, inverted=None,
                            logarithmic=None):
        self.execute("SetPlotAxisAttrs", {"axis": axis, "label": label, "autoRange": auto_range, "min": min, "max": max,
                                          "inverted": inverted, "logarithmic": logarithmic})

    def add_text(self, plot, x, y, label, color):
        """
            color: string, color name or r,g,b -  e.g "red" or "255.0.0"
        """
        return self.execute("AddText", {"plot": plot, "x": x, "y": y, "label": label, "color": color})

    def add_marker(self, axis, val, label, color):
        """
            color: see addText
        """
        return self.execute("AddMarker", {"axis": axis, "val": val, "label": label, "color": color})

    def add_interval_marker(self, axis, start, end, label, color):
        """
            color: see addText
        """
        return self.execute("AddIntervalMarker",
                            {"axis": axis, "start": start, "end": end, "label": label, "color": color})

    def set_line_series_data(self, series, x, y, error=None, error_y=None):
        """
            x, y, error and error_y are arrays of floats.
            Error arguments is disregarded if plot type is not error.
        """
        self.execute("SetLineSeriesData", {"series": series, "x": x, "y": y, "error": error, "errorY": error_y})

    def set_matrix_series_data(self, series, data, x=None, y=None):
        """
            x, y: array of loats
            data: array of arrays
        """
        self.execute("SetMatrixSeriesData", {"series": series, "x": x, "y": y, "data": data})

    def add_line_series(self, plot, name, color=None, axis_y=None, marker_size=None, line_width=None, max_count=None):
        return self.execute("AddLineSeries",
                            {"plot": plot, "name": name, "color": color, "axisY": axis_y, "markerSize": marker_size,
                             "lineWidth": line_width, "maxCount": max_count})

    def add_matrix_series(self, plot, name, min_x=None, max_x=None, bins_x=None, min_y=None, max_y=None, bins_y=None):
        return self.execute("AddMatrixSeries",
                            {"plot": plot, "name": name, "minX": min_x, "maxX": max_x, "nX": bins_x, "minY": min_y,
                             "maxY": max_y, "nY": bins_y})

    def add_3d_series(self, plot, name, min_x=None, max_x=None, bins_x=None, min_y=None, max_y=None, bins_y=None,
                      min_z=None, max_z=None, bins_z=None):
        return self.execute("Add3dSeries",
                            {"plot": plot, "name": name, "minX": min_x, "maxX": max_x, "nX": bins_x, "minY": min_y,
                             "maxY": max_y, "nY": bins_y, "minZ": min_z, "maxZ": max_z, "nZ": bins_z})

    def add_time_series(self, plot, name, color=None, axis_y=None):
        return self.execute("AddTimeSeries", {"plot": plot, "name": name, "color": color, "axisY": axis_y})

    def set_line_series_attrs(self, series, color=None, marker_size=None, line_width=None, max_count=None):
        self.execute("SetLineSeriesAttrs",
                     {"series": series, "color": color, "markerSize": marker_size, "lineWidth": line_width,
                      "maxCount": max_count})

    def set_matrix_series_attrs(self, series, min_x=None, max_x=None, bins_x=None, min_y=None, max_y=None, bins_y=None):
        self.execute("SetMatrixSeriesAttrs",
                     {"series": series, "minX": min_x, "maxX": max_x, "nX": bins_x, "minY": min_y, "maxY": max_y,
                      "nY": bins_y})

    def set_time_series_attrs(self, series, color):
        self.execute("SetTimeSeriesAttrs", {"series": series, "color": color})

    def append_line_series_data(self, series, x, y, error=None):
        """
        x, y, error and error are floats.
        Error argument is disregarded if plot type is not error.
        """
        self.execute("AppendLineSeriesData", {"series": series, "x": x, "y": y, "error": error})

    def append_line_series_data_array(self, series, x, y, error=None):
        """
            x, y, error are arrays of floats.
            Error argument is disregarded if plot type is not error.
        """
        self.execute("AppendLineSeriesDataArray", {"series": series, "x": x, "y": y, "error": error})

    def append_matrix_series_data(self, series, x, y, z):
        """
            x, y, and z are floats.
        """
        self.execute("AppendMatrixSeriesData", {"series": series, "x": x, "y": y, "z": z})

    def append_matrix_series_data_array(self, series, x, y, z):
        """
            x, y, and z arrays of floats.
        """
        self.execute("AppendMatrixSeriesDataArray", {"series": series, "x": x, "y": y, "z": z})

    def append_3d_series_data(self, series, data):
        """
        """
        self.execute("Append3dSeriesData", {"series": series, "data": data})

    def append_time_series_data(self, series, time=None, value=None):
        """
            x, y, and z are floats.
        """
        self.execute("AppendTimeSeriesData", {"series": series, "time": time, "value": value})

    def remove_series(self, series):
        self.execute("RemoveSeries", {"series": series})

    def remove_marker(self, marker):
        self.execute("RemoveMarker", {"marker": marker})

    def remove_text(self, text):
        self.execute("RemoveText", {"text": text})

    def get_plot_snapshot(self, plot, type="png", width=None, height=None):
        """
            type (str): png, gif, bmp ot jpg
        """
        ret = self.execute("GetPlotSnapshot", {"plot": plot, "type": type, "width": width, "height": height})
        ret = json.loads(ret)
        ret = base64.b64decode(ret)
        return ret




import time
import unittest
from datahub import *

url = "tcp://localhost:9999"
mode = "PULL"
filename = "/Users/gobbo_a/dev/back/bsread.h5"
channels = None
channels = ["UInt8Scalar", "Float32Scalar"]
start = 0.0
end = 2.1
query = {
    "channels": channels,
    "start": start,
    "end": end,
    "filter": "UInt8Scalar<100"
}

class BsreadTest(unittest.TestCase):
    def setUp(self):
        self.source = Bsread(url=url, mode=mode, time_type="str")

    def tearDown(self):
        self.source.close()


    def test_listeners(self):
        #hdf5 = HDF5Writer(filename, default_compression=Compression.GZIP)
        stdout = Stdout()
        table = Table()
        #self.source.set_id("bsread")
        #self.source.add_listener(hdf5)
        self.source.add_listener(stdout)
        self.source.add_listener(table)
        self.source.request(query)
        self.source.req(channels, start, end, receive_timeout=5000, filter= "UInt8Scalar>100")
        dataframe = table.as_dataframe()
        print(dataframe.columns)
        if channels:
            self.assertEqual(list(dataframe.keys()), channels)
        self.source.close_listeners()

    def test_bsread_stream(self):
        with BsreadStream(url=url, mode=mode,time_type="str", channels=channels, filter= "UInt8Scalar<10") as source:
            for i in range(10):
                print(i, source.receive(1.0))

    def test_bsread_no_channels(self):
        stdout = Stdout()
        self.source.add_listener(stdout)
        self.source.req(None, 0.0, 1.0)

if __name__ == '__main__':
    unittest.main()
